/*!
 * @brief 特殊属性武器で攻撃した際の追加効果処理
 * @date 2020/05/23
 * @author Hourier
 * @details
 * カオス属性、魔術属性、ゴールデンハンマー
 */

#include "player-attack/attack-chaos-effect.h"
#include "artifact/fixed-art-types.h"
#include "flavor/flavor-describer.h"
#include "flavor/object-flavor-types.h"
#include "inventory/inventory-object.h"
#include "inventory/inventory-slot-types.h"
#include "monster-race//race-ability-mask.h"
#include "monster-race/race-flags-resistance.h"
#include "monster-race/race-resistance-mask.h"
#include "monster/monster-describer.h"
#include "monster/monster-info.h"
#include "monster/monster-status-setter.h"
#include "monster/monster-status.h"
#include "object/object-mark-types.h"
#include "player-attack/player-attack.h"
#include "player/attack-defense-types.h"
#include "realm/realm-hex-numbers.h"
#include "spell-kind/spells-polymorph.h"
#include "spell-kind/spells-sight.h"
#include "spell-kind/spells-teleport.h"
#include "spell-realm/spells-hex.h"
#include "system/floor/floor-info.h"
#include "system/item-entity.h"
#include "system/monrace/monrace-definition.h"
#include "system/monrace/monrace-list.h"
#include "system/monster-entity.h"
#include "system/player-type-definition.h"
#include "system/redrawing-flags-updater.h"
#include "util/bit-flags-calculator.h"
#include "util/string-processor.h"
#include "view/display-messages.h"

/*!
 * @brief 打撃でモンスターを混乱させる処理
 * @param player_ptr プレイヤーへの参照ポインタ
 * @param pa_ptr 直接攻撃構造体への参照ポインタ
 * @param can_resist レベルで抵抗可能ならTRUE、できないならFALSE
 * @details
 * カオス属性や混乱の手
 */
static void attack_confuse(PlayerType *player_ptr, player_attack_type *pa_ptr, bool can_resist = true)
{
    if (player_ptr->special_attack & ATTACK_CONFUSE) {
        player_ptr->special_attack &= ~(ATTACK_CONFUSE);
        msg_print(_("手の輝きがなくなった。", "Your hands stop glowing."));
        RedrawingFlagsUpdater::get_instance().set_flag(MainWindowRedrawingFlag::TIMED_EFFECT);
    }

    auto &monrace = *pa_ptr->r_ptr;
    if (monrace.resistance_flags.has(MonsterResistanceType::NO_CONF)) {
        if (is_original_ap_and_seen(player_ptr, *pa_ptr->m_ptr)) {
            monrace.r_resistance_flags.set(MonsterResistanceType::NO_CONF);
        }
        msg_format(_("%s^には効果がなかった。", "%s^ is unaffected."), pa_ptr->m_name);

    } else if (can_resist && evaluate_percent(monrace.level)) {
        msg_format(_("%s^には効果がなかった。", "%s^ is unaffected."), pa_ptr->m_name);
    } else {
        msg_format(_("%s^は混乱したようだ。", "%s^ appears confused."), pa_ptr->m_name);
        (void)set_monster_confused(player_ptr, pa_ptr->m_idx, pa_ptr->m_ptr->get_remaining_confusion() + 10 + randint0(player_ptr->lev) / 5);
    }
}

/*!
 * @brief 打撃でモンスターを朦朧とさせる処理
 * @param player_ptr プレイヤーへの参照ポインタ
 * @param pa_ptr 直接攻撃構造体への参照ポインタ
 * @param can_resist レベルで抵抗可能ならTRUE、できないならFALSE
 * @details
 * 魔術属性
 */
static void attack_stun(PlayerType *player_ptr, player_attack_type *pa_ptr, bool can_resist = true)
{
    auto &monrace = *pa_ptr->r_ptr;
    if (monrace.resistance_flags.has(MonsterResistanceType::NO_STUN)) {
        if (is_original_ap_and_seen(player_ptr, *pa_ptr->m_ptr)) {
            monrace.resistance_flags.set(MonsterResistanceType::NO_STUN);
        }
        msg_format(_("%s^には効果がなかった。", "%s^ is unaffected."), pa_ptr->m_name);
    } else if (can_resist && evaluate_percent(monrace.level)) {
        msg_format(_("%s^には効果がなかった。", "%s^ is unaffected."), pa_ptr->m_name);
    } else {
        msg_format(_("%s^は朦朧としたようだ。", "%s^ appears stunned."), pa_ptr->m_name);
        (void)set_monster_stunned(player_ptr, pa_ptr->m_idx, pa_ptr->m_ptr->get_remaining_stun() + 10 + randint0(player_ptr->lev) / 5);
    }
}

/*!
 * @brief 打撃でモンスターを恐怖させる処理
 * @param player_ptr プレイヤーへの参照ポインタ
 * @param pa_ptr 直接攻撃構造体への参照ポインタ
 * @param can_resist レベルで抵抗可能ならTRUE、できないならFALSE
 * @details
 * 魔術属性
 */
static void attack_scare(PlayerType *player_ptr, player_attack_type *pa_ptr, bool can_resist = true)
{
    auto &monrace = *pa_ptr->r_ptr;
    if (monrace.resistance_flags.has(MonsterResistanceType::NO_FEAR)) {
        if (is_original_ap_and_seen(player_ptr, *pa_ptr->m_ptr)) {
            monrace.resistance_flags.set(MonsterResistanceType::NO_FEAR);
        }
        msg_format(_("%s^には効果がなかった。", "%s^ is unaffected."), pa_ptr->m_name);
    } else if (can_resist && evaluate_percent(monrace.level)) {
        msg_format(_("%s^には効果がなかった。", "%s^ is unaffected."), pa_ptr->m_name);
    } else {
        msg_format(_("%s^は恐怖して逃げ出した！", "%s^ flees in terror!"), pa_ptr->m_name);
        (void)set_monster_monfear(player_ptr, pa_ptr->m_idx, pa_ptr->m_ptr->get_remaining_fear() + 10 + randint0(player_ptr->lev) / 5);
    }
}

/*!
 * @brief 打撃でモンスターを無力化する処理
 * @param player_ptr プレイヤーへの参照ポインタ
 * @param pa_ptr 直接攻撃構造体への参照ポインタ
 * @details
 * 魔術属性
 */
static void attack_dispel(PlayerType *player_ptr, player_attack_type *pa_ptr)
{
    if (pa_ptr->r_ptr->ability_flags.has_none_of(RF_ABILITY_ATTACK_MASK) && pa_ptr->r_ptr->ability_flags.has_none_of(RF_ABILITY_INDIRECT_MASK)) {
        return;
    }

    auto dd = 2;
    if (pa_ptr->m_ptr->mtimed[MonsterTimedEffect::SLOW]) {
        dd += 1;
    }
    if (pa_ptr->m_ptr->mtimed[MonsterTimedEffect::FAST]) {
        dd += 2;
    }
    if (pa_ptr->m_ptr->mtimed[MonsterTimedEffect::INVULNERABILITY]) {
        dd += 3;
    }

    msg_print(_("武器が敵の魔力を吸い取った！", "The weapon drains mana from your enemy!"));
    dispel_monster_status(player_ptr, pa_ptr->m_idx);

    auto sp = Dice::roll(dd, 8);
    player_ptr->csp = std::min(player_ptr->msp, player_ptr->csp + sp);
    RedrawingFlagsUpdater::get_instance().set_flag(MainWindowRedrawingFlag::MP);
}

/*!
 * @brief 打撃でモンスターを調査する処理
 * @param player_ptr プレイヤーへの参照ポインタ
 * @param pa_ptr 直接攻撃構造体への参照ポインタ
 * @details
 * 魔術属性
 */
static void attack_probe(PlayerType *player_ptr, player_attack_type *pa_ptr)
{
    msg_print(_("刃が敵を調査した...", "The blade probed your enemy..."));
    msg_print(nullptr);
    msg_print(probed_monster_info(player_ptr, *pa_ptr->m_ptr, *pa_ptr->r_ptr));
    msg_print(nullptr);
    const auto mes = MonraceList::get_instance().probe_lore(pa_ptr->r_idx);
    if (mes) {
        msg_print(*mes);
        msg_print(nullptr);
    }
}

/*!
 * @breif カオス武器でのテレポート・アウェイを行うか判定する (抵抗されたら無効)
 * @param player_ptr プレイヤーへの参照ポインタ
 * @param pa_ptr 直接攻撃構造体への参照ポインタ
 * @return 抵抗されたらTRUE、アウェイされるならFALSE
 */
static bool judge_tereprt_resistance(PlayerType *player_ptr, player_attack_type *pa_ptr)
{
    auto &monrace = *pa_ptr->r_ptr;
    if (monrace.resistance_flags.has_not(MonsterResistanceType::RESIST_TELEPORT)) {
        return false;
    }

    if (monrace.kind_flags.has(MonsterKindType::UNIQUE)) {
        if (is_original_ap_and_seen(player_ptr, *pa_ptr->m_ptr)) {
            monrace.r_resistance_flags.set(MonsterResistanceType::RESIST_TELEPORT);
        }

        msg_format(_("%s^には効果がなかった。", "%s^ is unaffected!"), pa_ptr->m_name);
        return true;
    }

    if (monrace.level > randint1(100)) {
        if (is_original_ap_and_seen(player_ptr, *pa_ptr->m_ptr)) {
            monrace.r_resistance_flags.set(MonsterResistanceType::RESIST_TELEPORT);
        }

        msg_format(_("%s^は抵抗力を持っている！", "%s^ resists!"), pa_ptr->m_name);
        return true;
    }

    return false;
}

/*!
 * @brief カオス武器でのテレポート・アウェイを実行する
 * @param player_ptr プレイヤーへの参照ポインタ
 * @param pa_ptr 直接攻撃構造体への参照ポインタ
 * @param num 現在の攻撃回数 (テレポートしてしまったら追加攻撃できないのでその補正)
 */
static void attack_teleport_away(PlayerType *player_ptr, player_attack_type *pa_ptr, int *num)
{
    if (judge_tereprt_resistance(player_ptr, pa_ptr)) {
        return;
    }

    msg_format(_("%s^は消えた！", "%s^ disappears!"), pa_ptr->m_name);
    teleport_away(player_ptr, pa_ptr->m_idx, 50, TELEPORT_PASSIVE);
    *num = pa_ptr->num_blow + 1;
    *(pa_ptr->mdeath) = true;
}

/*!
 * @brief カオス武器でのテレポート・アウェイを実行する
 * @param player_ptr プレイヤーへの参照ポインタ
 * @param pa_ptr 直接攻撃構造体への参照ポインタ
 * @param y モンスターのY座標
 * @param x モンスターのX座標
 */
static void attack_polymorph(PlayerType *player_ptr, player_attack_type *pa_ptr, POSITION y, POSITION x)
{
    const auto &monrace = *pa_ptr->r_ptr;
    if (monrace.kind_flags.has(MonsterKindType::UNIQUE) || monrace.misc_flags.has(MonsterMiscType::QUESTOR) || monrace.resistance_flags.has_any_of(RFR_EFF_RESIST_CHAOS_MASK)) {
        return;
    }

    if (polymorph_monster(player_ptr, y, x)) {
        msg_format(_("%s^は変化した！", "%s^ changes!"), pa_ptr->m_name);
        *(pa_ptr->fear) = false;
        pa_ptr->weak = false;
    } else {
        msg_format(_("%s^には効果がなかった。", "%s^ is unaffected."), pa_ptr->m_name);
    }

    pa_ptr->m_ptr = &player_ptr->current_floor_ptr->m_list[pa_ptr->m_idx];
    angband_strcpy(pa_ptr->m_name, monster_desc(player_ptr, *pa_ptr->m_ptr, 0), sizeof(pa_ptr->m_name));
}

/*!
 * @brief ゴールデンハンマーによるアイテム奪取処理
 * @param player_ptr プレイヤーへの参照ポインタ
 * @param pa_ptr 直接攻撃構造体への参照ポインタ
 */
static void attack_golden_hammer(PlayerType *player_ptr, player_attack_type *pa_ptr)
{
    auto &floor = *player_ptr->current_floor_ptr;
    auto &monster = floor.m_list[pa_ptr->m_idx];
    if (monster.hold_o_idx_list.empty()) {
        return;
    }

    auto &item = *floor.o_list[monster.hold_o_idx_list.front()];
    const auto item_name = describe_flavor(player_ptr, item, OD_NAME_ONLY);
    item.held_m_idx = 0;
    item.marked.clear().set(OmType::TOUCHED);
    monster.hold_o_idx_list.pop_front();
    msg_format(_("%sを奪った。", "You snatched %s."), item_name.data());
    store_item_to_inventory(player_ptr, &item);
}

/*!
 * @brief カオス武器その他でモンスターのステータスを変化させる
 * @param player_ptr プレイヤーへの参照ポインタ
 * @param pa_ptr 直接攻撃構造体への参照ポインタ
 * @param y モンスターのY座標
 * @param x モンスターのX座標
 * @param num 現在の攻撃回数
 */
void change_monster_stat(PlayerType *player_ptr, player_attack_type *pa_ptr, const POSITION y, const POSITION x, int *num)
{
    auto should_confuse = any_bits(player_ptr->special_attack, ATTACK_CONFUSE);
    should_confuse |= pa_ptr->chaos_effect == CE_CONFUSION;
    should_confuse |= pa_ptr->mode == HISSATSU_CONF;
    should_confuse |= SpellHex(player_ptr).is_spelling_specific(HEX_CONFUSION);
    if (should_confuse) {
        attack_confuse(player_ptr, pa_ptr);
    }

    if (pa_ptr->magical_effect == MagicalBrandEffectType::STUN) {
        attack_stun(player_ptr, pa_ptr, false);
    }

    if (pa_ptr->magical_effect == MagicalBrandEffectType::SCARE) {
        attack_scare(player_ptr, pa_ptr, false);
    }

    if (pa_ptr->magical_effect == MagicalBrandEffectType::DISPELL) {
        attack_dispel(player_ptr, pa_ptr);
    }

    if (pa_ptr->magical_effect == MagicalBrandEffectType::PROBE) {
        attack_probe(player_ptr, pa_ptr);
    }

    if (pa_ptr->chaos_effect == CE_TELE_AWAY) {
        attack_teleport_away(player_ptr, pa_ptr, num);
    }

    if (pa_ptr->chaos_effect == CE_POLYMORPH && (randint1(90) > pa_ptr->m_ptr->get_monrace().level)) {
        attack_polymorph(player_ptr, pa_ptr, y, x);
    }

    const auto &item = *player_ptr->inventory[enum2i(INVEN_MAIN_HAND) + pa_ptr->hand];
    if (item.is_specific_artifact(FixedArtifactId::G_HAMMER)) {
        attack_golden_hammer(player_ptr, pa_ptr);
    }
}
