#include "smith/smith-tables.h"
#include "artifact/random-art-effects.h"
#include "object-enchant/tr-flags.h"
#include "object-enchant/tr-types.h"
#include "smith/object-smith.h"
#include "smith/smith-info.h"
#include "smith/smith-types.h"

#include <memory>
#include <unordered_map>
#include <vector>

/*!
 * @brief エッセンスの順序リスト。エッセンスの表示順等で使用する。
 */
const std::vector<SmithEssenceType> Smith::essence_list_order = {
    SmithEssenceType::STR,
    SmithEssenceType::INT,
    SmithEssenceType::WIS,
    SmithEssenceType::DEX,
    SmithEssenceType::CON,
    SmithEssenceType::CHR,

    SmithEssenceType::SUST_STATUS,

    SmithEssenceType::MAGIC_MASTERY,
    SmithEssenceType::STEALTH,
    SmithEssenceType::SEARCH,
    SmithEssenceType::INFRA,
    SmithEssenceType::TUNNEL,
    SmithEssenceType::SPEED,
    SmithEssenceType::BLOWS,

    SmithEssenceType::CHAOTIC,
    SmithEssenceType::VAMPIRIC,
    SmithEssenceType::EATHQUAKE,
    SmithEssenceType::BRAND_POIS,
    SmithEssenceType::BRAND_ACID,
    SmithEssenceType::BRAND_ELEC,
    SmithEssenceType::BRAND_FIRE,
    SmithEssenceType::BRAND_COLD,
    SmithEssenceType::BRAND_MAGIC,

    SmithEssenceType::IMMUNITY,
    SmithEssenceType::REFLECT,

    SmithEssenceType::RES_ACID,
    SmithEssenceType::RES_ELEC,
    SmithEssenceType::RES_FIRE,
    SmithEssenceType::RES_COLD,
    SmithEssenceType::RES_POIS,
    SmithEssenceType::RES_FEAR,
    SmithEssenceType::RES_LITE,
    SmithEssenceType::RES_DARK,
    SmithEssenceType::RES_BLIND,
    SmithEssenceType::RES_CONF,
    SmithEssenceType::RES_SOUND,
    SmithEssenceType::RES_SHARDS,
    SmithEssenceType::RES_NETHER,
    SmithEssenceType::RES_NEXUS,
    SmithEssenceType::RES_CHAOS,
    SmithEssenceType::RES_DISEN,
    SmithEssenceType::RES_WATER,
    SmithEssenceType::RES_TIME,
    SmithEssenceType::RES_CURSE,

    SmithEssenceType::HOLD_EXP,
    SmithEssenceType::FREE_ACT,
    SmithEssenceType::WARNING,
    SmithEssenceType::LEVITATION,
    SmithEssenceType::SEE_INVIS,
    SmithEssenceType::SLOW_DIGEST,
    SmithEssenceType::REGEN,
    SmithEssenceType::TELEPORT,
    SmithEssenceType::NO_MAGIC,
    SmithEssenceType::LITE,
    SmithEssenceType::NO_TELE,
    SmithEssenceType::TELEPATHY,

    SmithEssenceType::SLAY_EVIL,
    SmithEssenceType::SLAY_ANIMAL,
    SmithEssenceType::SLAY_UNDEAD,
    SmithEssenceType::SLAY_DEMON,
    SmithEssenceType::SLAY_ORC,
    SmithEssenceType::SLAY_TROLL,
    SmithEssenceType::SLAY_GIANT,
    SmithEssenceType::SLAY_DRAGON,
    SmithEssenceType::SLAY_HUMAN,
    SmithEssenceType::SLAY_GOOD,

    SmithEssenceType::EASY2_WEAPON,
    SmithEssenceType::STRENGTHEN_BOW,
    SmithEssenceType::UNIQUE,

    SmithEssenceType::ATTACK,
    SmithEssenceType::AC,
};

/*!
 * @brief SmithEssence からエッセンスの表記名を引く連想配列
 */
const std::unordered_map<SmithEssenceType, concptr> Smith::essence_to_name = {
    { SmithEssenceType::STR, _("腕力", "strength") },
    { SmithEssenceType::INT, _("知能", "intelligen.") },
    { SmithEssenceType::WIS, _("賢さ", "wisdom") },
    { SmithEssenceType::DEX, _("器用さ", "dexterity") },
    { SmithEssenceType::CON, _("耐久力", "constitut.") },
    { SmithEssenceType::CHR, _("魅力", "charisma") },

    { SmithEssenceType::SUST_STATUS, _("能力維持", "sustain") },

    { SmithEssenceType::MAGIC_MASTERY, _("魔力支配", "magic mast.") },
    { SmithEssenceType::STEALTH, _("隠密", "stealth") },
    { SmithEssenceType::SEARCH, _("探索", "searching") },
    { SmithEssenceType::INFRA, _("赤外線視力", "infravision") },
    { SmithEssenceType::TUNNEL, _("採掘", "digging") },
    { SmithEssenceType::SPEED, _("スピード", "speed") },
    { SmithEssenceType::BLOWS, _("追加攻撃", "extra atk") },

    { SmithEssenceType::CHAOTIC, _("カオス攻撃", "chaos brand") },
    { SmithEssenceType::VAMPIRIC, _("吸血攻撃", "vampiric") },
    { SmithEssenceType::EATHQUAKE, _("地震", "quake") },
    { SmithEssenceType::BRAND_POIS, _("毒殺", "pois. brand") },
    { SmithEssenceType::BRAND_ACID, _("溶解", "acid brand") },
    { SmithEssenceType::BRAND_ELEC, _("電撃", "elec. brand") },
    { SmithEssenceType::BRAND_FIRE, _("焼棄", "fire brand") },
    { SmithEssenceType::BRAND_COLD, _("凍結", "cold brand") },
    { SmithEssenceType::BRAND_MAGIC, _("魔術", "magic brand") },

    { SmithEssenceType::IMMUNITY, _("免疫", "immunity") },
    { SmithEssenceType::REFLECT, _("反射", "reflection") },

    { SmithEssenceType::RES_ACID, _("耐酸", "res. acid") },
    { SmithEssenceType::RES_ELEC, _("耐電撃", "res. elec.") },
    { SmithEssenceType::RES_FIRE, _("耐火炎", "res. fire") },
    { SmithEssenceType::RES_COLD, _("耐冷気", "res. cold") },
    { SmithEssenceType::RES_POIS, _("耐毒", "res. poison") },
    { SmithEssenceType::RES_FEAR, _("耐恐怖", "res. fear") },
    { SmithEssenceType::RES_LITE, _("耐閃光", "res. light") },
    { SmithEssenceType::RES_DARK, _("耐暗黒", "res. dark") },
    { SmithEssenceType::RES_BLIND, _("耐盲目", "res. blind") },
    { SmithEssenceType::RES_CONF, _("耐混乱", "res.confuse") },
    { SmithEssenceType::RES_SOUND, _("耐轟音", "res. sound") },
    { SmithEssenceType::RES_SHARDS, _("耐破片", "res. shard") },
    { SmithEssenceType::RES_NETHER, _("耐地獄", "res. nether") },
    { SmithEssenceType::RES_NEXUS, _("耐因果混乱", "res. nexus") },
    { SmithEssenceType::RES_CHAOS, _("耐カオス", "res. chaos") },
    { SmithEssenceType::RES_DISEN, _("耐劣化", "res. disen.") },
    { SmithEssenceType::RES_WATER, _("耐水", "res. water") },
    { SmithEssenceType::RES_TIME, _("耐時間逆転", "res. time") },
    { SmithEssenceType::RES_CURSE, _("耐呪力", "res. curse") },

    { SmithEssenceType::HOLD_EXP, _("経験値維持", "hold exp") },
    { SmithEssenceType::FREE_ACT, _("麻痺知らず", "free action") },
    { SmithEssenceType::WARNING, _("警告", "warning") },
    { SmithEssenceType::LEVITATION, _("浮遊", "levitation") },
    { SmithEssenceType::SEE_INVIS, _("可視透明", "see invis.") },
    { SmithEssenceType::TELEPATHY, _("テレパシー", "telepathy") },
    { SmithEssenceType::SLOW_DIGEST, _("遅消化", "slow dige.") },
    { SmithEssenceType::REGEN, _("急速回復", "regen.") },
    { SmithEssenceType::TELEPORT, _("テレポート", "teleport") },
    { SmithEssenceType::NO_MAGIC, _("反魔法", "anti magic") },
    { SmithEssenceType::LITE, _("永久光源", "perm. light") },
    { SmithEssenceType::NO_TELE, _("反テレポ", "no teleport") },

    { SmithEssenceType::SLAY_EVIL, _("邪悪倍打", "slay evil") },
    { SmithEssenceType::SLAY_ANIMAL, _("動物倍打", "slay animal") },
    { SmithEssenceType::SLAY_UNDEAD, _("不死倍打", "slay undead") },
    { SmithEssenceType::SLAY_DEMON, _("悪魔倍打", "slay demon") },
    { SmithEssenceType::SLAY_ORC, _("オーク倍打", "slay orc") },
    { SmithEssenceType::SLAY_TROLL, _("トロル倍打", "slay troll") },
    { SmithEssenceType::SLAY_GIANT, _("巨人倍打", "slay giant") },
    { SmithEssenceType::SLAY_DRAGON, _("竜倍打", "slay dragon") },
    { SmithEssenceType::SLAY_HUMAN, _("人間倍打", "slay human") },
    { SmithEssenceType::SLAY_GOOD, _("善良倍打", "slay good") },

    { SmithEssenceType::EASY2_WEAPON, _("二刀流", "two weapons") },
    { SmithEssenceType::STRENGTHEN_BOW, _("射撃強化", "enc. ranged") },
    { SmithEssenceType::UNIQUE, _("ユニーク", "unique") },

    { SmithEssenceType::ATTACK, _("攻撃", "weapon enc.") },
    { SmithEssenceType::AC, _("防御", "armor enc.") },
};

/*!
 * @brief エッセンス抽出情報テーブル
 */
const std::vector<essence_drain_type> Smith::essence_drain_info_table = {
    { TR_STR, { SmithEssenceType::STR }, 10 },
    { TR_INT, { SmithEssenceType::INT }, 10 },
    { TR_WIS, { SmithEssenceType::WIS }, 10 },
    { TR_DEX, { SmithEssenceType::DEX }, 10 },
    { TR_CON, { SmithEssenceType::CON }, 10 },
    { TR_CHR, { SmithEssenceType::CHR }, 10 },
    { TR_MAGIC_MASTERY, { SmithEssenceType::MAGIC_MASTERY }, 10 },
    { TR_FORCE_WEAPON, { SmithEssenceType::INT, SmithEssenceType::WIS }, 5 },
    { TR_STEALTH, { SmithEssenceType::STEALTH }, 10 },
    { TR_SEARCH, { SmithEssenceType::SEARCH }, 10 },
    { TR_INFRA, { SmithEssenceType::INFRA }, 10 },
    { TR_TUNNEL, { SmithEssenceType::TUNNEL }, 10 },
    { TR_SPEED, { SmithEssenceType::SPEED }, 10 },
    { TR_BLOWS, { SmithEssenceType::BLOWS }, 10 },
    { TR_CHAOTIC, { SmithEssenceType::CHAOTIC }, 10 },
    { TR_VAMPIRIC, { SmithEssenceType::VAMPIRIC }, 10 },
    { TR_SLAY_ANIMAL, { SmithEssenceType::SLAY_ANIMAL }, 10 },
    { TR_SLAY_EVIL, { SmithEssenceType::SLAY_EVIL }, 10 },
    { TR_SLAY_UNDEAD, { SmithEssenceType::SLAY_UNDEAD }, 10 },
    { TR_SLAY_DEMON, { SmithEssenceType::SLAY_DEMON }, 10 },
    { TR_SLAY_ORC, { SmithEssenceType::SLAY_ORC }, 10 },
    { TR_SLAY_TROLL, { SmithEssenceType::SLAY_TROLL }, 10 },
    { TR_SLAY_GIANT, { SmithEssenceType::SLAY_GIANT }, 10 },
    { TR_SLAY_DRAGON, { SmithEssenceType::SLAY_DRAGON }, 10 },
    { TR_KILL_DRAGON, { SmithEssenceType::SLAY_DRAGON }, 10 },
    { TR_VORPAL, { SmithEssenceType::BRAND_POIS, SmithEssenceType::BRAND_ACID, SmithEssenceType::BRAND_ELEC, SmithEssenceType::BRAND_FIRE, SmithEssenceType::BRAND_COLD }, 5 },
    { TR_EARTHQUAKE, { SmithEssenceType::EATHQUAKE }, 10 },
    { TR_BRAND_POIS, { SmithEssenceType::BRAND_POIS }, 10 },
    { TR_BRAND_ACID, { SmithEssenceType::BRAND_ACID }, 10 },
    { TR_BRAND_ELEC, { SmithEssenceType::BRAND_ELEC }, 10 },
    { TR_BRAND_FIRE, { SmithEssenceType::BRAND_FIRE }, 10 },
    { TR_BRAND_COLD, { SmithEssenceType::BRAND_COLD }, 10 },

    { TR_SUST_STR, { SmithEssenceType::SUST_STATUS }, 10 },
    { TR_SUST_INT, { SmithEssenceType::SUST_STATUS }, 10 },
    { TR_SUST_WIS, { SmithEssenceType::SUST_STATUS }, 10 },
    { TR_SUST_DEX, { SmithEssenceType::SUST_STATUS }, 10 },
    { TR_SUST_CON, { SmithEssenceType::SUST_STATUS }, 10 },
    { TR_SUST_CHR, { SmithEssenceType::SUST_STATUS }, 10 },
    { TR_RIDING, {}, 0 },
    { TR_EASY_SPELL, { SmithEssenceType::BRAND_MAGIC }, 10 },
    { TR_IM_ACID, { SmithEssenceType::IMMUNITY }, 100 },
    { TR_IM_ELEC, { SmithEssenceType::IMMUNITY }, 100 },
    { TR_IM_FIRE, { SmithEssenceType::IMMUNITY }, 100 },
    { TR_IM_COLD, { SmithEssenceType::IMMUNITY }, 100 },
    { TR_THROW, {}, 0 },
    { TR_REFLECT, { SmithEssenceType::REFLECT }, 10 },
    { TR_FREE_ACT, { SmithEssenceType::FREE_ACT }, 10 },
    { TR_HOLD_EXP, { SmithEssenceType::HOLD_EXP }, 10 },
    { TR_RES_ACID, { SmithEssenceType::RES_ACID }, 10 },
    { TR_RES_ELEC, { SmithEssenceType::RES_ELEC }, 10 },
    { TR_RES_FIRE, { SmithEssenceType::RES_FIRE }, 10 },
    { TR_RES_COLD, { SmithEssenceType::RES_COLD }, 10 },
    { TR_RES_POIS, { SmithEssenceType::RES_POIS }, 10 },
    { TR_RES_FEAR, { SmithEssenceType::RES_FEAR }, 10 },
    { TR_RES_LITE, { SmithEssenceType::RES_LITE }, 10 },
    { TR_RES_DARK, { SmithEssenceType::RES_DARK }, 10 },
    { TR_RES_BLIND, { SmithEssenceType::RES_BLIND }, 10 },
    { TR_RES_CONF, { SmithEssenceType::RES_CONF }, 10 },
    { TR_RES_SOUND, { SmithEssenceType::RES_SOUND }, 10 },
    { TR_RES_SHARDS, { SmithEssenceType::RES_SHARDS }, 10 },
    { TR_RES_NETHER, { SmithEssenceType::RES_NETHER }, 10 },
    { TR_RES_NEXUS, { SmithEssenceType::RES_NEXUS }, 10 },
    { TR_RES_CHAOS, { SmithEssenceType::RES_CHAOS }, 10 },
    { TR_RES_DISEN, { SmithEssenceType::RES_DISEN }, 10 },

    { TR_SH_FIRE, { SmithEssenceType::BRAND_FIRE, SmithEssenceType::RES_FIRE }, 10 },
    { TR_SH_ELEC, { SmithEssenceType::BRAND_ELEC, SmithEssenceType::RES_ELEC }, 10 },
    { TR_SLAY_HUMAN, { SmithEssenceType::SLAY_HUMAN }, 10 },
    { TR_SH_COLD, { SmithEssenceType::BRAND_COLD, SmithEssenceType::RES_COLD }, 10 },
    { TR_NO_TELE, { SmithEssenceType::NO_TELE }, 10 },
    { TR_NO_MAGIC, { SmithEssenceType::NO_MAGIC }, 10 },
    { TR_DEC_MANA, { SmithEssenceType::INT }, 10 },
    { TR_TY_CURSE, {}, -1 },
    { TR_WARNING, { SmithEssenceType::WARNING }, 10 },
    { TR_HIDE_TYPE, {}, 0 },
    { TR_SHOW_MODS, {}, 0 },
    { TR_SLAY_GOOD, { SmithEssenceType::SLAY_GOOD }, 10 },
    { TR_LEVITATION, { SmithEssenceType::LEVITATION }, 10 },
    { TR_LITE_1, { SmithEssenceType::LITE }, 10 },
    { TR_SEE_INVIS, { SmithEssenceType::SEE_INVIS }, 10 },
    { TR_TELEPATHY, { SmithEssenceType::TELEPATHY }, 10 },
    { TR_SLOW_DIGEST, { SmithEssenceType::SLOW_DIGEST }, 10 },
    { TR_REGEN, { SmithEssenceType::REGEN }, 10 },
    { TR_XTRA_MIGHT, { SmithEssenceType::STRENGTHEN_BOW }, 10 },
    { TR_XTRA_SHOTS, { SmithEssenceType::STRENGTHEN_BOW }, 10 },
    { TR_IGNORE_ACID, {}, 0 },
    { TR_IGNORE_ELEC, {}, 0 },
    { TR_IGNORE_FIRE, {}, 0 },
    { TR_IGNORE_COLD, {}, 0 },
    { TR_ACTIVATE, {}, 0 },
    { TR_DRAIN_EXP, {}, -1 },
    { TR_TELEPORT, { SmithEssenceType::TELEPORT }, 10 },
    { TR_AGGRAVATE, {}, -1 },
    { TR_BLESSED, {}, 0 },
    { TR_XXX_93, {}, 0 },
    { TR_XXX_94, {}, 0 },
    { TR_KILL_GOOD, { SmithEssenceType::SLAY_GOOD }, 10 },

    { TR_KILL_ANIMAL, { SmithEssenceType::SLAY_ANIMAL }, 10 },
    { TR_KILL_EVIL, { SmithEssenceType::SLAY_EVIL }, 10 },
    { TR_KILL_UNDEAD, { SmithEssenceType::SLAY_UNDEAD }, 10 },
    { TR_KILL_DEMON, { SmithEssenceType::SLAY_DEMON }, 10 },
    { TR_KILL_ORC, { SmithEssenceType::SLAY_ORC }, 10 },
    { TR_KILL_TROLL, { SmithEssenceType::SLAY_TROLL }, 10 },
    { TR_KILL_GIANT, { SmithEssenceType::SLAY_GIANT }, 10 },
    { TR_KILL_HUMAN, { SmithEssenceType::SLAY_HUMAN }, 10 },
    { TR_ESP_ANIMAL, { SmithEssenceType::SLAY_ANIMAL }, 10 },
    { TR_ESP_UNDEAD, { SmithEssenceType::SLAY_UNDEAD }, 10 },
    { TR_ESP_DEMON, { SmithEssenceType::SLAY_DEMON }, 10 },
    { TR_ESP_ORC, { SmithEssenceType::SLAY_ORC }, 10 },
    { TR_ESP_TROLL, { SmithEssenceType::SLAY_TROLL }, 10 },
    { TR_ESP_GIANT, { SmithEssenceType::SLAY_GIANT }, 10 },
    { TR_ESP_DRAGON, { SmithEssenceType::SLAY_DRAGON }, 10 },
    { TR_ESP_HUMAN, { SmithEssenceType::SLAY_HUMAN }, 10 },
    { TR_ESP_EVIL, { SmithEssenceType::SLAY_EVIL }, 10 },
    { TR_ESP_GOOD, { SmithEssenceType::SLAY_GOOD }, 10 },
    { TR_ESP_NONLIVING, {}, 0 }, // TODO
    { TR_ESP_UNIQUE, { SmithEssenceType::UNIQUE }, 20 },
    { TR_FULL_NAME, {}, 0 },
    { TR_FIXED_FLAVOR, {}, 0 },
    { TR_ADD_L_CURSE, {}, -1 },
    { TR_ADD_H_CURSE, {}, -1 },
    { TR_DRAIN_HP, {}, -1 },
    { TR_DRAIN_MANA, {}, -1 },
    { TR_LITE_2, { SmithEssenceType::LITE }, 20 },
    { TR_LITE_3, { SmithEssenceType::LITE }, 30 },
    { TR_LITE_M1, {}, 0 },
    { TR_LITE_M2, {}, 0 },
    { TR_LITE_M3, {}, 0 },
    { TR_LITE_FUEL, {}, 0 },

    { TR_CALL_ANIMAL, {}, -1 },
    { TR_CALL_DEMON, {}, -1 },
    { TR_CALL_DRAGON, {}, -1 },
    { TR_CALL_UNDEAD, {}, -1 },
    { TR_COWARDICE, {}, -1 },
    { TR_LOW_MELEE, {}, -1 },
    { TR_LOW_AC, {}, -1 },
    { TR_HARD_SPELL, {}, -1 },
    { TR_FAST_DIGEST, {}, -1 },
    { TR_SLOW_REGEN, {}, -1 },
    { TR_MIGHTY_THROW, {}, 0 },
    { TR_EASY2_WEAPON, { SmithEssenceType::EASY2_WEAPON }, 10 },
    { TR_DOWN_SAVING, {}, -1 },
    { TR_NO_AC, {}, -1 },
    { TR_XXX_142, {}, -1 },
    { TR_RES_TIME, { SmithEssenceType::RES_TIME }, 10 },
    { TR_RES_WATER, { SmithEssenceType::RES_WATER }, 10 },
    { TR_INVULN_ARROW, {}, 0 },
    { TR_DARK_SOURCE, {}, 0 },
    { TR_SUPPORTIVE, { SmithEssenceType::EASY2_WEAPON }, 5 },
    { TR_RES_CURSE, { SmithEssenceType::RES_CURSE }, 10 },
    { TR_BERS_RAGE, {}, -1 },
    { TR_BRAND_MAGIC, { SmithEssenceType::BRAND_MAGIC }, 10 },
    { TR_IMPACT, {}, 0 },
    { TR_VUL_ACID, {}, -1 },
    { TR_VUL_COLD, {}, -1 },
    { TR_VUL_ELEC, {}, -1 },
    { TR_VUL_FIRE, {}, -1 },
    { TR_VUL_LITE, {}, -1 },
    { TR_IM_DARK, {}, 0 },

    { TR_SELF_FIRE, { SmithEssenceType::BRAND_FIRE, SmithEssenceType::RES_FIRE }, 10 },
    { TR_SELF_ELEC, { SmithEssenceType::BRAND_ELEC, SmithEssenceType::RES_ELEC }, 10 },
    { TR_SELF_COLD, { SmithEssenceType::BRAND_COLD, SmithEssenceType::RES_COLD }, 10 },
    { TR_PERSISTENT_CURSE, {}, -1 },
    { TR_VUL_CURSE, {}, -1 },
};

namespace {

template <typename T, typename... Args>
std::shared_ptr<ISmithInfo> make_info(SmithEffectType effect, concptr name, SmithCategoryType category, std::vector<SmithEssenceType> need_essences, int consumption, Args &&...args)
{
    return std::make_shared<T>(effect, name, category, std::move(need_essences), consumption, std::forward<Args>(args)...);
}

std::shared_ptr<ISmithInfo> make_basic_smith_info(SmithEffectType effect, concptr name, SmithCategoryType category, std::vector<SmithEssenceType> need_essences, int consumption, TrFlags add_flags)
{
    return make_info<BasicSmithInfo>(effect, name, category, std::move(need_essences), consumption, std::move(add_flags));
}

}

/*!
 * @brief 鍛冶情報テーブル
 */
const std::vector<std::shared_ptr<ISmithInfo>> Smith::smith_info_table = {
    make_basic_smith_info(SmithEffectType::NONE, _("なし", "None"), SmithCategoryType::NONE, std::vector<SmithEssenceType>{ SmithEssenceType::NONE }, 0, {}),

    make_basic_smith_info(SmithEffectType::STR, _("腕力", "strength"), SmithCategoryType::PVAL, { SmithEssenceType::STR }, 20, { TR_STR }),
    make_basic_smith_info(SmithEffectType::INT, _("知能", "intelligence"), SmithCategoryType::PVAL, { SmithEssenceType::INT }, 20, { TR_INT }),
    make_basic_smith_info(SmithEffectType::WIS, _("賢さ", "wisdom"), SmithCategoryType::PVAL, { SmithEssenceType::WIS }, 20, { TR_WIS }),
    make_basic_smith_info(SmithEffectType::DEX, _("器用さ", "dexterity"), SmithCategoryType::PVAL, { SmithEssenceType::DEX }, 20, { TR_DEX }),
    make_basic_smith_info(SmithEffectType::CON, _("耐久力", "constitution"), SmithCategoryType::PVAL, { SmithEssenceType::CON }, 20, { TR_CON }),
    make_basic_smith_info(SmithEffectType::CHR, _("魅力", "charisma"), SmithCategoryType::PVAL, { SmithEssenceType::CHR }, 20, { TR_CHR }),

    make_basic_smith_info(SmithEffectType::SUST_STR, _("腕力維持", "sustain strength"), SmithCategoryType::ABILITY, { SmithEssenceType::SUST_STATUS }, 15, { TR_SUST_STR }),
    make_basic_smith_info(SmithEffectType::SUST_INT, _("知能維持", "sustain intelligence"), SmithCategoryType::ABILITY, { SmithEssenceType::SUST_STATUS }, 15, { TR_SUST_INT }),
    make_basic_smith_info(SmithEffectType::SUST_WIS, _("賢さ維持", "sustain wisdom"), SmithCategoryType::ABILITY, { SmithEssenceType::SUST_STATUS }, 15, { TR_SUST_WIS }),
    make_basic_smith_info(SmithEffectType::SUST_DEX, _("器用さ維持", "sustain dexterity"), SmithCategoryType::ABILITY, { SmithEssenceType::SUST_STATUS }, 15, { TR_SUST_DEX }),
    make_basic_smith_info(SmithEffectType::SUST_CON, _("耐久力維持", "sustain constitution"), SmithCategoryType::ABILITY, { SmithEssenceType::SUST_STATUS }, 15, { TR_SUST_CON }),
    make_basic_smith_info(SmithEffectType::SUST_CHR, _("魅力維持", "sustain charisma"), SmithCategoryType::ABILITY, { SmithEssenceType::SUST_STATUS }, 15, { TR_SUST_CHR }),

    make_basic_smith_info(SmithEffectType::MAGIC_MASTERY, _("魔力支配", "magic mastery"), SmithCategoryType::PVAL, { SmithEssenceType::MAGIC_MASTERY }, 20, { TR_MAGIC_MASTERY }),
    make_basic_smith_info(SmithEffectType::STEALTH, _("隠密", "stealth"), SmithCategoryType::PVAL, { SmithEssenceType::STEALTH }, 40, { TR_STEALTH }),
    make_basic_smith_info(SmithEffectType::SEARCH, _("探索", "searching"), SmithCategoryType::PVAL, { SmithEssenceType::SEARCH }, 15, { TR_SEARCH }),
    make_basic_smith_info(SmithEffectType::INFRA, _("赤外線視力", "infravision"), SmithCategoryType::PVAL, { SmithEssenceType::INFRA }, 15, { TR_INFRA }),
    make_basic_smith_info(SmithEffectType::TUNNEL, _("採掘", "digging"), SmithCategoryType::PVAL, { SmithEssenceType::TUNNEL }, 15, { TR_TUNNEL }),
    make_basic_smith_info(SmithEffectType::SPEED, _("スピード", "speed"), SmithCategoryType::PVAL, { SmithEssenceType::SPEED }, 12, { TR_SPEED }),
    make_basic_smith_info(SmithEffectType::BLOWS, _("追加攻撃", "extra attack"), SmithCategoryType::WEAPON_ATTR, { SmithEssenceType::BLOWS }, 20, { TR_BLOWS }),

    make_basic_smith_info(SmithEffectType::CHAOTIC, _("カオス攻撃", "chaos brand"), SmithCategoryType::WEAPON_ATTR, { SmithEssenceType::CHAOTIC }, 15, { TR_CHAOTIC }),
    make_basic_smith_info(SmithEffectType::VAMPIRIC, _("吸血攻撃", "vampiric brand"), SmithCategoryType::WEAPON_ATTR, { SmithEssenceType::VAMPIRIC }, 60, { TR_VAMPIRIC }),
    make_basic_smith_info(SmithEffectType::EARTHQUAKE, _("地震攻撃", "quake brand"), SmithCategoryType::WEAPON_ATTR, { SmithEssenceType::EATHQUAKE }, 15, { TR_EARTHQUAKE }),
    make_basic_smith_info(SmithEffectType::BRAND_POIS, _("毒殺", "poison brand"), SmithCategoryType::WEAPON_ATTR, { SmithEssenceType::BRAND_POIS }, 20, { TR_BRAND_POIS }),
    make_basic_smith_info(SmithEffectType::BRAND_ACID, _("溶解", "acid brand"), SmithCategoryType::WEAPON_ATTR, { SmithEssenceType::BRAND_ACID }, 20, { TR_BRAND_ACID }),
    make_basic_smith_info(SmithEffectType::BRAND_ELEC, _("電撃", "electric brand"), SmithCategoryType::WEAPON_ATTR, { SmithEssenceType::BRAND_ELEC }, 20, { TR_BRAND_ELEC }),
    make_basic_smith_info(SmithEffectType::BRAND_FIRE, _("焼棄", "fire brand"), SmithCategoryType::WEAPON_ATTR, { SmithEssenceType::BRAND_FIRE }, 20, { TR_BRAND_FIRE }),
    make_basic_smith_info(SmithEffectType::BRAND_COLD, _("凍結", "cold brand"), SmithCategoryType::WEAPON_ATTR, { SmithEssenceType::BRAND_COLD }, 20, { TR_BRAND_COLD }),
    make_basic_smith_info(SmithEffectType::VORPAL, _("切れ味", "sharpness"), SmithCategoryType::WEAPON_ATTR, { SmithEssenceType::BRAND_ACID, SmithEssenceType::BRAND_ELEC, SmithEssenceType::BRAND_FIRE, SmithEssenceType::BRAND_COLD }, 10, { TR_VORPAL }),
    make_basic_smith_info(SmithEffectType::BRAND_MAGIC, _("魔術属性攻撃", "magic brand"), SmithCategoryType::WEAPON_ATTR, { SmithEssenceType::BRAND_MAGIC }, 20, { TR_BRAND_MAGIC }),
    make_basic_smith_info(SmithEffectType::XTRA_MIGHT, _("射撃倍率強化", "enhance firing power"), SmithCategoryType::WEAPON_ATTR, { SmithEssenceType::STRENGTHEN_BOW }, 50, { TR_XTRA_MIGHT }),
    make_basic_smith_info(SmithEffectType::XTRA_SHOTS, _("射撃速度強化", "enhance firing rate"), SmithCategoryType::WEAPON_ATTR, { SmithEssenceType::STRENGTHEN_BOW }, 50, { TR_XTRA_SHOTS }),

    make_basic_smith_info(SmithEffectType::IM_ACID, _("酸免疫", "acid immunity"), SmithCategoryType::RESISTANCE, { SmithEssenceType::IMMUNITY, SmithEssenceType::RES_ACID }, 200, { TR_IM_ACID }),
    make_basic_smith_info(SmithEffectType::IM_ELEC, _("電撃免疫", "electric immunity"), SmithCategoryType::RESISTANCE, { SmithEssenceType::IMMUNITY, SmithEssenceType::RES_ELEC }, 200, { TR_IM_ELEC }),
    make_basic_smith_info(SmithEffectType::IM_FIRE, _("火炎免疫", "fire immunity"), SmithCategoryType::RESISTANCE, { SmithEssenceType::IMMUNITY, SmithEssenceType::RES_FIRE }, 200, { TR_IM_FIRE }),
    make_basic_smith_info(SmithEffectType::IM_COLD, _("冷気免疫", "cold immunity"), SmithCategoryType::RESISTANCE, { SmithEssenceType::IMMUNITY, SmithEssenceType::RES_COLD }, 200, { TR_IM_COLD }),
    make_basic_smith_info(SmithEffectType::REFLECT, _("反射", "reflection"), SmithCategoryType::RESISTANCE, { SmithEssenceType::REFLECT }, 20, { TR_REFLECT }),

    make_basic_smith_info(SmithEffectType::RES_ACID, _("耐酸", "resistance to acid"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_ACID }, 15, { TR_RES_ACID }),
    make_basic_smith_info(SmithEffectType::RES_ELEC, _("耐電撃", "resistance to electric"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_ELEC }, 15, { TR_RES_ELEC }),
    make_basic_smith_info(SmithEffectType::RES_FIRE, _("耐火炎", "resistance to fire"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_FIRE }, 15, { TR_RES_FIRE }),
    make_basic_smith_info(SmithEffectType::RES_COLD, _("耐冷気", "resistance to cold"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_COLD }, 15, { TR_RES_COLD }),
    make_basic_smith_info(SmithEffectType::RES_POIS, _("耐毒", "resistance to poison"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_POIS }, 25, { TR_RES_POIS }),
    make_basic_smith_info(SmithEffectType::RES_FEAR, _("耐恐怖", "resistance to fear"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_FEAR }, 20, { TR_RES_FEAR }),
    make_basic_smith_info(SmithEffectType::RES_LITE, _("耐閃光", "resistance to light"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_LITE }, 20, { TR_RES_LITE }),
    make_basic_smith_info(SmithEffectType::RES_DARK, _("耐暗黒", "resistance to dark"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_DARK }, 20, { TR_RES_DARK }),
    make_basic_smith_info(SmithEffectType::RES_BLIND, _("耐盲目", "resistance to blind"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_BLIND }, 20, { TR_RES_BLIND }),
    make_basic_smith_info(SmithEffectType::RES_CONF, _("耐混乱", "resistance to confusion"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_CONF }, 20, { TR_RES_CONF }),
    make_basic_smith_info(SmithEffectType::RES_SOUND, _("耐轟音", "resistance to sound"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_SOUND }, 20, { TR_RES_SOUND }),
    make_basic_smith_info(SmithEffectType::RES_SHARDS, _("耐破片", "resistance to shard"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_SHARDS }, 20, { TR_RES_SHARDS }),
    make_basic_smith_info(SmithEffectType::RES_NETHER, _("耐地獄", "resistance to nether"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_NETHER }, 20, { TR_RES_NETHER }),
    make_basic_smith_info(SmithEffectType::RES_NEXUS, _("耐因果混乱", "resistance to nexus"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_NEXUS }, 20, { TR_RES_NEXUS }),
    make_basic_smith_info(SmithEffectType::RES_CHAOS, _("耐カオス", "resistance to chaos"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_CHAOS }, 20, { TR_RES_CHAOS }),
    make_basic_smith_info(SmithEffectType::RES_DISEN, _("耐劣化", "resistance to disenchantment"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_DISEN }, 20, { TR_RES_DISEN }),
    make_basic_smith_info(SmithEffectType::RES_WATER, _("耐水", "resistance to water"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_WATER }, 20, { TR_RES_WATER }),
    make_basic_smith_info(SmithEffectType::RES_TIME, _("耐時間逆転", "resistance to time"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_TIME }, 20, { TR_RES_TIME }),
    make_basic_smith_info(SmithEffectType::RES_CURSE, _("耐呪力", "resistance to curse"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_CURSE }, 20, { TR_RES_CURSE }),

    make_basic_smith_info(SmithEffectType::HOLD_EXP, _("経験値維持", "hold experience"), SmithCategoryType::ABILITY, { SmithEssenceType::HOLD_EXP }, 20, { TR_HOLD_EXP }),
    make_basic_smith_info(SmithEffectType::FREE_ACT, _("麻痺知らず", "free action"), SmithCategoryType::ABILITY, { SmithEssenceType::FREE_ACT }, 20, { TR_FREE_ACT }),
    make_basic_smith_info(SmithEffectType::WARNING, _("警告", "warning"), SmithCategoryType::ABILITY, { SmithEssenceType::WARNING }, 20, { TR_WARNING }),
    make_basic_smith_info(SmithEffectType::LEVITATION, _("浮遊", "levitation"), SmithCategoryType::ABILITY, { SmithEssenceType::LEVITATION }, 20, { TR_LEVITATION }),
    make_basic_smith_info(SmithEffectType::SEE_INVIS, _("可視透明", "see invisible"), SmithCategoryType::ABILITY, { SmithEssenceType::SEE_INVIS }, 20, { TR_SEE_INVIS }),
    make_basic_smith_info(SmithEffectType::SLOW_DIGEST, _("遅消化", "slow digestion"), SmithCategoryType::ABILITY, { SmithEssenceType::SLOW_DIGEST }, 15, { TR_SLOW_DIGEST }),
    make_basic_smith_info(SmithEffectType::REGEN, _("急速回復", "regeneration"), SmithCategoryType::ABILITY, { SmithEssenceType::REGEN }, 20, { TR_REGEN }),
    make_basic_smith_info(SmithEffectType::TELEPORT, _("テレポート", "teleport"), SmithCategoryType::ABILITY, { SmithEssenceType::TELEPORT }, 25, { TR_TELEPORT }),
    make_basic_smith_info(SmithEffectType::NO_MAGIC, _("反魔法", "anti magic"), SmithCategoryType::ABILITY, { SmithEssenceType::NO_MAGIC }, 15, { TR_NO_MAGIC }),
    make_basic_smith_info(SmithEffectType::LITE, _("永久光源", "permanent light"), SmithCategoryType::ABILITY, { SmithEssenceType::LITE }, 15, { TR_LITE_1 }),
    make_basic_smith_info(SmithEffectType::NO_TELE, _("反テレポート", "prevent teleportation"), SmithCategoryType::ABILITY, { SmithEssenceType::NO_TELE }, 15, { TR_NO_TELE }),

    make_basic_smith_info(SmithEffectType::SLAY_EVIL, _("邪悪倍打", "slay evil"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_EVIL }, 100, { TR_SLAY_EVIL }),
    make_basic_smith_info(SmithEffectType::SLAY_ANIMAL, _("動物倍打", "slay animal"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_ANIMAL }, 20, { TR_SLAY_ANIMAL }),
    make_basic_smith_info(SmithEffectType::SLAY_UNDEAD, _("不死倍打", "slay undead"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_UNDEAD }, 20, { TR_SLAY_UNDEAD }),
    make_basic_smith_info(SmithEffectType::SLAY_DEMON, _("悪魔倍打", "slay demon"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_DEMON }, 20, { TR_SLAY_DEMON }),
    make_basic_smith_info(SmithEffectType::SLAY_ORC, _("オーク倍打", "slay orc"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_ORC }, 20, { TR_SLAY_ORC }),
    make_basic_smith_info(SmithEffectType::SLAY_TROLL, _("トロル倍打", "slay troll"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_TROLL }, 20, { TR_SLAY_TROLL }),
    make_basic_smith_info(SmithEffectType::SLAY_GIANT, _("巨人倍打", "slay giant"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_GIANT }, 20, { TR_SLAY_GIANT }),
    make_basic_smith_info(SmithEffectType::SLAY_DRAGON, _("竜倍打", "slay dragon"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_DRAGON }, 20, { TR_SLAY_DRAGON }),
    make_basic_smith_info(SmithEffectType::SLAY_HUMAN, _("人間倍打", "slay human"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_HUMAN }, 20, { TR_SLAY_HUMAN }),
    make_basic_smith_info(SmithEffectType::SLAY_GOOD, _("善良倍打", "slay good"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_GOOD }, 20, { TR_SLAY_GOOD }),

    make_basic_smith_info(SmithEffectType::KILL_EVIL, _("邪悪倍倍打", "kill evil"), SmithCategoryType::NONE, { SmithEssenceType::SLAY_EVIL }, 0, { TR_KILL_EVIL }), // 強力すぎるため無効(SmithCategory:NONE)
    make_basic_smith_info(SmithEffectType::KILL_ANIMAL, _("動物倍倍打", "kill animal"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_ANIMAL }, 60, { TR_KILL_ANIMAL }),
    make_basic_smith_info(SmithEffectType::KILL_UNDEAD, _("不死倍倍打", "kill undead"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_UNDEAD }, 60, { TR_KILL_UNDEAD }),
    make_basic_smith_info(SmithEffectType::KILL_DEMON, _("悪魔倍倍打", "kill demon"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_DEMON }, 60, { TR_KILL_DEMON }),
    make_basic_smith_info(SmithEffectType::KILL_ORC, _("オーク倍倍打", "kill orc"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_ORC }, 60, { TR_KILL_ORC }),
    make_basic_smith_info(SmithEffectType::KILL_TROLL, _("トロル倍倍打", "kill troll"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_TROLL }, 60, { TR_KILL_TROLL }),
    make_basic_smith_info(SmithEffectType::KILL_GIANT, _("巨人倍倍打", "kill giant"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_GIANT }, 60, { TR_KILL_GIANT }),
    make_basic_smith_info(SmithEffectType::KILL_DRAGON, _("竜倍倍打", "kill dragon"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_DRAGON }, 60, { TR_KILL_DRAGON }),
    make_basic_smith_info(SmithEffectType::KILL_HUMAN, _("人間倍倍打", "kill human"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_HUMAN }, 60, { TR_KILL_HUMAN }),
    make_basic_smith_info(SmithEffectType::KILL_GOOD, _("善良倍倍打", "kill good"), SmithCategoryType::SLAYING, { SmithEssenceType::SLAY_GOOD }, 60, { TR_KILL_GOOD }),

    make_basic_smith_info(SmithEffectType::TELEPATHY, _("テレパシー", "telepathy"), SmithCategoryType::ESP, { SmithEssenceType::TELEPATHY }, 15, { TR_TELEPATHY }),
    make_basic_smith_info(SmithEffectType::ESP_ANIMAL, _("動物ESP", "sense animal"), SmithCategoryType::ESP, { SmithEssenceType::SLAY_ANIMAL }, 40, { TR_ESP_ANIMAL }),
    make_basic_smith_info(SmithEffectType::ESP_UNDEAD, _("不死ESP", "sense undead"), SmithCategoryType::ESP, { SmithEssenceType::SLAY_UNDEAD }, 40, { TR_ESP_UNDEAD }),
    make_basic_smith_info(SmithEffectType::ESP_DEMON, _("悪魔ESP", "sense demon"), SmithCategoryType::ESP, { SmithEssenceType::SLAY_DEMON }, 40, { TR_ESP_DEMON }),
    make_basic_smith_info(SmithEffectType::ESP_ORC, _("オークESP", "sense orc"), SmithCategoryType::ESP, { SmithEssenceType::SLAY_ORC }, 40, { TR_ESP_ORC }),
    make_basic_smith_info(SmithEffectType::ESP_TROLL, _("トロルESP", "sense troll"), SmithCategoryType::ESP, { SmithEssenceType::SLAY_TROLL }, 40, { TR_ESP_TROLL }),
    make_basic_smith_info(SmithEffectType::ESP_GIANT, _("巨人ESP", "sense giant"), SmithCategoryType::ESP, { SmithEssenceType::SLAY_GIANT }, 40, { TR_ESP_GIANT }),
    make_basic_smith_info(SmithEffectType::ESP_DRAGON, _("竜ESP", "sense dragon"), SmithCategoryType::ESP, { SmithEssenceType::SLAY_DRAGON }, 40, { TR_ESP_DRAGON }),
    make_basic_smith_info(SmithEffectType::ESP_HUMAN, _("人間ESP", "sense human"), SmithCategoryType::ESP, { SmithEssenceType::SLAY_HUMAN }, 40, { TR_ESP_HUMAN }),
    make_basic_smith_info(SmithEffectType::ESP_GOOD, _("善良ESP", "sense good"), SmithCategoryType::ESP, { SmithEssenceType::SLAY_GOOD }, 40, { TR_ESP_GOOD }),
    make_basic_smith_info(SmithEffectType::ESP_UNIQUE, _("ユニークESP", "sense unique"), SmithCategoryType::ESP, { SmithEssenceType::UNIQUE }, 100, { TR_ESP_UNIQUE }),

    make_basic_smith_info(SmithEffectType::SH_FIRE, _("火炎オーラ", "fiery sheath"), SmithCategoryType::ETC, { SmithEssenceType::RES_FIRE, SmithEssenceType::BRAND_FIRE }, 50, { TR_RES_FIRE, TR_SH_FIRE }),
    make_basic_smith_info(SmithEffectType::SH_ELEC, _("電撃オーラ", "electric sheath"), SmithCategoryType::ETC, { SmithEssenceType::RES_ELEC, SmithEssenceType::BRAND_ELEC }, 50, { TR_RES_ELEC, TR_SH_ELEC }),
    make_basic_smith_info(SmithEffectType::SH_COLD, _("冷気オーラ", "sheath of coldness"), SmithCategoryType::ETC, { SmithEssenceType::RES_COLD, SmithEssenceType::BRAND_COLD }, 50, { TR_RES_COLD, TR_SH_COLD }),

    make_basic_smith_info(SmithEffectType::RESISTANCE, _("全耐性", "resistance"), SmithCategoryType::RESISTANCE, { SmithEssenceType::RES_ACID, SmithEssenceType::RES_ELEC, SmithEssenceType::RES_FIRE, SmithEssenceType::RES_COLD }, 150, { TR_RES_ACID, TR_RES_ELEC, TR_RES_FIRE, TR_RES_COLD }),
    make_info<SlayingGlovesSmithInfo>(SmithEffectType::SLAY_GLOVE, _("殺戮の小手", "gauntlets of slaying"), SmithCategoryType::ETC, { SmithEssenceType::ATTACK }, 200),
    make_basic_smith_info(SmithEffectType::EASY_2WEAPON, _("源氏の小手", "gauntlets of Genji"), SmithCategoryType::ETC, { SmithEssenceType::EASY2_WEAPON }, 20, { TR_EASY2_WEAPON }),

    make_info<ActivationSmithInfo>(SmithEffectType::ACT_QUAKE, _("地震", "quake"), SmithCategoryType::ACTIVATION, { SmithEssenceType::EATHQUAKE }, 15, RandomArtActType::QUAKE),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_RES_ACID, _("酸耐性", "resist acid"), SmithCategoryType::ACTIVATION, { SmithEssenceType::RES_ACID }, 30, RandomArtActType::RESIST_ACID),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_RES_ELEC, _("電撃耐性", "resist electricity"), SmithCategoryType::ACTIVATION, { SmithEssenceType::RES_ELEC }, 30, RandomArtActType::RESIST_ELEC),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_RES_FIRE, _("火炎耐性", "resist fire"), SmithCategoryType::ACTIVATION, { SmithEssenceType::RES_FIRE }, 30, RandomArtActType::RESIST_FIRE),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_RES_COLD, _("冷気耐性", "resist cold"), SmithCategoryType::ACTIVATION, { SmithEssenceType::RES_COLD }, 30, RandomArtActType::RESIST_COLD),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_RES_POIS, _("毒耐性", "resist poison"), SmithCategoryType::ACTIVATION, { SmithEssenceType::RES_POIS }, 40, RandomArtActType::RESIST_POIS),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_PHASE_DOOR, _("ショート・テレポート", "blink"), SmithCategoryType::ACTIVATION, { SmithEssenceType::TELEPORT }, 30, RandomArtActType::PHASE_DOOR),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_TELEPORT, _("テレポート", "teleport"), SmithCategoryType::ACTIVATION, { SmithEssenceType::TELEPORT }, 40, RandomArtActType::TELEPORT),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_SPEED, _("スピード", "speed"), SmithCategoryType::ACTIVATION, { SmithEssenceType::SPEED }, 25, RandomArtActType::SPEED),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_STONE_MUD, _("岩石溶解", "stone to mud"), SmithCategoryType::ACTIVATION, { SmithEssenceType::TUNNEL }, 100, RandomArtActType::STONE_MUD),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_LIGHT, _("イルミネーション", "light area"), SmithCategoryType::ACTIVATION, { SmithEssenceType::LITE }, 30, RandomArtActType::LIGHT),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_MAP_LIGHT, _("魔法の地図と光", "light & map area"), SmithCategoryType::ACTIVATION, { SmithEssenceType::SEARCH, SmithEssenceType::LITE }, 30, RandomArtActType::MAP_LIGHT),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_DETECT_ALL, _("全感知", "detection"), SmithCategoryType::ACTIVATION, { SmithEssenceType::SEARCH, SmithEssenceType::TELEPATHY }, 30, RandomArtActType::DETECT_ALL),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_DETECT_UNIQUE, _("階にいるユニークモンスターを表示", "list of the uniques on the level"), SmithCategoryType::ACTIVATION, { SmithEssenceType::UNIQUE }, 100, RandomArtActType::DETECT_UNIQUE),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_REST_EXP, _("経験値復活", "restore experience"), SmithCategoryType::ACTIVATION, { SmithEssenceType::HOLD_EXP }, 30, RandomArtActType::REST_EXP),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_REST_ALL, _("全ステータスと経験値復活", "restore stats and experience"), SmithCategoryType::ACTIVATION, { SmithEssenceType::SUST_STATUS, SmithEssenceType::HOLD_EXP }, 40, RandomArtActType::REST_ALL),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_PROT_EVIL, _("対邪悪結界", "protect evil"), SmithCategoryType::ACTIVATION, { SmithEssenceType::SLAY_EVIL }, 100, RandomArtActType::PROT_EVIL),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_DISP_EVIL, _("邪悪退散", "dispel evil"), SmithCategoryType::ACTIVATION, { SmithEssenceType::SLAY_EVIL }, 100, RandomArtActType::DISP_EVIL),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_DISP_GOOD, _("善良退散", "dispel good"), SmithCategoryType::ACTIVATION, { SmithEssenceType::SLAY_GOOD }, 50, RandomArtActType::DISP_GOOD),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_PESTICIDE, _("害虫駆除", "dispel pests"), SmithCategoryType::ACTIVATION, { SmithEssenceType::BRAND_POIS }, 30, RandomArtActType::PESTICIDE),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_BA_ACID, _("アシッド・ボール", "ball of acid"), SmithCategoryType::ACTIVATION, { SmithEssenceType::BRAND_ACID }, 40, RandomArtActType::BA_ACID_1),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_BA_ELEC, _("サンダー・ボール", "ball of lightning"), SmithCategoryType::ACTIVATION, { SmithEssenceType::BRAND_ELEC }, 40, RandomArtActType::BA_ELEC_2),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_BA_FIRE, _("ファイア・ボール", "ball of fire"), SmithCategoryType::ACTIVATION, { SmithEssenceType::BRAND_FIRE }, 40, RandomArtActType::BA_FIRE_4),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_BA_COLD, _("アイス・ボール", "ball of cold"), SmithCategoryType::ACTIVATION, { SmithEssenceType::BRAND_COLD }, 40, RandomArtActType::BA_COLD_2),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_BA_NUKE, _("放射能球", "ball of nuke"), SmithCategoryType::ACTIVATION, { SmithEssenceType::BRAND_POIS, SmithEssenceType::CHAOTIC }, 30, RandomArtActType::BA_NUKE_1),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_SUNLIGHT, _("太陽光線", "beam of sunlight"), SmithCategoryType::ACTIVATION, { SmithEssenceType::LITE }, 40, RandomArtActType::SUNLIGHT),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_DRAIN, _("吸血の矢", "drain bolt"), SmithCategoryType::ACTIVATION, { SmithEssenceType::VAMPIRIC }, 100, RandomArtActType::DRAIN_1),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_CONFUSE, _("パニック・モンスター", "confuse monster"), SmithCategoryType::ACTIVATION, { SmithEssenceType::CHAOTIC }, 30, RandomArtActType::CONFUSE),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_SATIATE, _("空腹充足", "satisfy hunger"), SmithCategoryType::ACTIVATION, { SmithEssenceType::SLOW_DIGEST }, 50, RandomArtActType::SATIATE),
    make_info<ActivationSmithInfo>(SmithEffectType::ACT_CURE_700, _("体力回復", "heal"), SmithCategoryType::ACTIVATION, { SmithEssenceType::REGEN }, 100, RandomArtActType::CURE_700),

    make_info<EnchantWeaponSmithInfo>(SmithEffectType::ATTACK, _("攻撃", "weapon enchant"), SmithCategoryType::ENCHANT, { SmithEssenceType::ATTACK }, 30),
    make_info<EnchantArmourSmithInfo>(SmithEffectType::AC, _("防御", "armor enchant"), SmithCategoryType::ENCHANT, { SmithEssenceType::AC }, 15),
    make_info<SustainSmithInfo>(SmithEffectType::SUSTAIN, _("装備保持", "elements proof"), SmithCategoryType::ENCHANT, { SmithEssenceType::RES_ACID, SmithEssenceType::RES_ELEC, SmithEssenceType::RES_FIRE, SmithEssenceType::RES_COLD }, 10),
};
